#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <time.h>
#include <ctype.h>

#include "fw.h"

#define FW_WINDOW_DEFAULT_WIDTH   512
#define FW_WINDOW_DEFAULT_HEIGHT  512
#define FW_WINDOW_DEFAULT_POS_X   0
#define FW_WINDOW_DEFAULT_POS_Y   0
#define FW_WINDOW_DEFAULT_TITLE   "OpenGL Application"

static void fw_application_finalize(void);

static inline float fw_get_time(void)
{
	return ((float)(clock()) / ((float)(CLOCKS_PER_SEC)));
}

static fw_global_data g_global;

const fw_application_data * fw_application(void)
{
	return &(g_global.application);
}

const fw_window_data * fw_window(void)
{
	return &(g_global.window);
}

const fw_input_data * fw_input(void)
{
	return &(g_global.input);
}

const fw_keyboard_data * fw_keyboard(void)
{
	return &(g_global.input.keyboard);
}

const fw_mouse_data * fw_mouse(void)
{
	return &(g_global.input.mouse);
}

const fw_time_data * fw_time(void)
{
	return &(g_global.time);
}

const fw_global_data * fw_global(void)
{
	return &(g_global);
}

void fw_application_quit(void)
{
	fw_application_finalize();

	exit(0);
}

static void fw_application_initialize(void)
{
	memset(&g_global, 0, sizeof(g_global));

	g_global.window.width  = FW_WINDOW_DEFAULT_WIDTH;
	g_global.window.height = FW_WINDOW_DEFAULT_HEIGHT;

	g_global.time.now = fw_get_time();

	glewInit();

	usr_application_initialize();
}

static void fw_application_finalize(void)
{
	usr_application_finalize();

	memset(&g_global, 0, sizeof(g_global));
}

static void fw_application_idle(void)
{
	const float now = fw_get_time();

	g_global.time.frame_delta = now - g_global.time.now;
	g_global.time.now = now;

	usr_application_idle();

	glutPostRedisplay();
}

void fw_window_set_title(const char * title)
{
	glutSetWindowTitle(title);
}

void fw_window_resize(int width, int height)
{
	glutReshapeWindow(width, height);
}

static void fw_key_pressed(unsigned char key, int x, int y)
{
	const int up_key = toupper((int)key);
	const int lo_key = tolower((int)key);

	(void)x;
	(void)y;

	g_global.input.keyboard.key_pressed[up_key] = FW_TRUE;
	g_global.input.keyboard.key_pressed[lo_key] = FW_TRUE;

	const unsigned char val = !(g_global.input.keyboard.key_toggle[up_key]);
	g_global.input.keyboard.key_toggle[up_key] = val;
	g_global.input.keyboard.key_toggle[lo_key] = val;

	usr_key_pressed((int)key);
}

static void fw_key_released(unsigned char key, int x, int y)
{
	const int up_key = toupper((int)key);
	const int lo_key = tolower((int)key);

	(void)x;
	(void)y;

	g_global.input.keyboard.key_pressed[up_key] = FW_FALSE;
	g_global.input.keyboard.key_pressed[lo_key] = FW_FALSE;

	usr_key_released((int)key);
}

static void fw_mouse_cursor_moved(int x, int y)
{
	y = g_global.window.height - 1 - y;

	g_global.input.mouse.previous_cursor_position[0] = g_global.input.mouse.cursor_position[0];
	g_global.input.mouse.previous_cursor_position[1] = g_global.input.mouse.cursor_position[1];

	g_global.input.mouse.cursor_position[0] = x;
	g_global.input.mouse.cursor_position[1] = y;

	g_global.input.mouse.cursor_position_delta[0] = g_global.input.mouse.cursor_position[0] - g_global.input.mouse.previous_cursor_position[0];
	g_global.input.mouse.cursor_position_delta[1] = g_global.input.mouse.cursor_position[1] - g_global.input.mouse.previous_cursor_position[1];

	usr_mouse_cursor_moved(x, y);
}

static void fw_mouse_button(int button, int state, int x, int y)
{
	int btn = 0;
	unsigned char pressed = FW_FALSE;

	y = g_global.window.height - 1 - y;

	switch (button)
	{
		case GLUT_LEFT_BUTTON:
			btn = 0;
		break;

		case GLUT_RIGHT_BUTTON:
			btn = 1;
		break;

		case GLUT_MIDDLE_BUTTON:
			btn = 2;
		break;

		default:
			return;
		break;
	}

	switch (state)
	{
		case GLUT_DOWN:
			pressed = FW_TRUE;
		break;

		case GLUT_UP:
			pressed = FW_FALSE;
		break;

		default:
			return;
		break;
	}

	g_global.input.mouse.button_pressed[btn] = pressed;

	if (pressed != 0)
	{
		usr_mouse_button_pressed(btn, x, y);
	}
	else
	{
		usr_mouse_button_released(btn, x, y);
	}
}

static void fw_window_reshape(int width, int height)
{
	g_global.window.width  = width;
	g_global.window.height = height;

	usr_window_resize(width, height);
}

static void fw_window_draw(void)
{
	usr_window_draw();

	glutSwapBuffers();
}

int main(int argc, char ** argv)
{
	printf("OpenGL Application Framework\n\n");

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_STENCIL | GLUT_DOUBLE);

	glutInitWindowSize(FW_WINDOW_DEFAULT_WIDTH, FW_WINDOW_DEFAULT_HEIGHT);
	glutInitWindowPosition(FW_WINDOW_DEFAULT_POS_X, FW_WINDOW_DEFAULT_POS_Y);
	glutCreateWindow(FW_WINDOW_DEFAULT_TITLE);

	glutKeyboardFunc(fw_key_pressed);
	glutKeyboardUpFunc(fw_key_released);
	glutMotionFunc(fw_mouse_cursor_moved);
	glutPassiveMotionFunc(fw_mouse_cursor_moved);
	glutMouseFunc(fw_mouse_button);
	glutReshapeFunc(fw_window_reshape);
	glutDisplayFunc(fw_window_draw);
	glutIdleFunc(fw_application_idle);

	fw_application_initialize();

	glutMainLoop();

	return 0;
}
